/*  Project: Open Source Gmini Tools
    file: IDEController.cpp, IDE controller emulation

    Copyright (C) 2004  The gmEmu Project <gmemu@donat.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include "Emulator.h"
#include <assert.h>
#include "util.h"
#include "Console.h"

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

#define ATA_DATA        (0xD0)
#define ATA_ERROR       (0xD2) 
#define ATA_NSECTOR     (0xD4)
#define ATA_SECTOR      (0xD6)
#define ATA_LCYL        (0xD8)
#define ATA_HCYL        (0xDA)
#define ATA_SELECT      (0xDC)
#define ATA_COMMAND     (0xDE)
#define ATA_CONTROL     (0xEC)
#define ATA_UNKNOWN     (0xF0)


#define ATA_FEATURE     ATA_ERROR
#define ATA_STATUS      ATA_COMMAND
#define ATA_ALT_STATUS  ATA_CONTROL

#define SELECT_DEVICE1  0x10
#define SELECT_LBA      0x40

#define STATUS_BSY      0x80
#define STATUS_RDY      0x40
#define STATUS_DF       0x20
#define STATUS_DRQ      0x08
#define STATUS_ERR      0x01

#define ERROR_ABRT      0x04

#define CONTROL_nIEN    0x02
#define CONTROL_SRST    0x04

#define CMD_READ_SECTORS           0x20
#define CMD_WRITE_SECTORS          0x30
#define CMD_READ_MULTIPLE          0xC4
#define CMD_WRITE_MULTIPLE         0xC5
#define CMD_SET_MULTIPLE_MODE      0xC6
#define CMD_STANDBY_IMMEDIATE      0xE0
#define CMD_STANDBY                0xE2
#define CMD_IDENTIFY               0xEC
#define CMD_SLEEP                  0xE6
#define CMD_SET_FEATURES           0xEF
#define CMD_SECURITY_FREEZE_LOCK   0xF5

#define PARTITION_OFFSET (64)

// FIXME: put in ide controller
static u8 disk_header[PARTITION_OFFSET][0x200];

//FIXME: This is non-portable code.
static void setByte( u8 * buffer, int offset, u8 value ) {
   buffer[offset] = value;
}

static void setWord( u8 * buffer, int offset, u16 value ) {
   *( u16 * ) ( buffer + offset ) = value;
}

static void setDWord( u8 * buffer, int offset, u32 value ) {
   *( u32 * ) ( buffer + offset ) = value;
}


IDEController::IDEController( Emulator * gm, const char* hdFileName ) {
   gmini = gm;
   pendingDataCount = 0;
   pendingSectors = 0;

   for( int i = 0; i < REG_END; i++ ) {
      registers[i] = 0;
   }

   for( int i = 0; i < 0x100; i++ ) {
      configData[i] = 0;        // I don't know it; thus fill with 0
   }
   configData[47] = 0x10; // sectors per ata request



   ///////////
   // Load partition image
   hdWrite = true;
   //hdWrite = strncmp("/dev", hdFileName, 4);
   // disable write if real device. 
#ifdef __CYGWIN__
   hdfd = open(hdFileName, (hdWrite ? O_RDWR : O_RDONLY), 0);
#else   
   hdfd = open(hdFileName, O_LARGEFILE | (hdWrite ? O_RDWR : O_RDONLY), 0);
#endif
   if (hdfd == -1)
      console.output( "ERROR: cannot open HD-image file.\n" );

   ///////////
   // Initialize boot sector
   u8* boot_sec = disk_header[0];
     //off_t fsize = hdFile.tellp(  );
     off_t fsize = lseek(hdfd, 0, SEEK_END);
     u32 nsec = fsize >> 9; // using a division doesn't seem to do it.
     console.output("HD-image file size = 0x%llx (0x%x sectors)\n", fsize, nsec);
      // MBR<
      for( int i = 0; i < 0x200; i++ ) {
         boot_sec[i] = 0;
      }
      const int partition1 = 0x1BE;
      setByte( boot_sec, partition1 + 0x00, 0x80 );  // bootable
      setByte( boot_sec, partition1 + 0x04, 0x0C );  // FAT32 LBA
      setDWord( boot_sec, partition1 + 0x08, PARTITION_OFFSET ); // start sector
      setDWord( boot_sec, partition1 + 0x0C, nsec  ); // #sectors
      setWord( boot_sec, 0x1FE, 0xAA55 );   // MBR sig
}

IDEController::~IDEController(  ) {
  close(hdfd);
}


void IDEController::readSector( int sector, u8 * buffer ) {
   if( sector < PARTITION_OFFSET ) {
      memcpy(buffer, disk_header[sector], 0x200);
   }
   else {
      int partSector = sector - PARTITION_OFFSET;
      lseek(hdfd, ( off_t )partSector * 0x200, SEEK_SET);
      int rc = read(hdfd, reinterpret_cast < char *>( buffer ), 0x200 );
      if (rc != 0x200) {
	perror("IDEController::readSector");
	exit(-1);
      }
   }
   if (TRACE_IDE_DATA) {
     gmini->tracer.log("Read sector 0x%x (%d)", sector, sector);
     gmini->tracer.logBuffer(buffer, 0x200);
   }
}

void IDEController::writeSector( int sector, u8 * buffer ) {
    assert(hdWrite);
    if (sector < PARTITION_OFFSET) {
      memcpy(disk_header[sector], buffer, 0x200);
    } else {
      int rc;
      sector -= PARTITION_OFFSET;      
      lseek(hdfd, ( off_t )sector * 0x200, SEEK_SET);
      rc = write(hdfd, reinterpret_cast < char *>( buffer ), 0x200);
      if (rc != 0x200) {
        perror("IDEController::writeSector");
	exit(-1);
      }
    }
    if (TRACE_IDE_DATA) {
      gmini->tracer.log("Written sector 0x%x (%d)", sector, sector);
      gmini->tracer.logBuffer(buffer, 0x200);
    }
}

void IDEController::transferData( int direction, u32 reg, u8 * memBuff, int numBytes ) {
   if (reg == ATA_UNKNOWN) {
     assert( direction == 0 );
     // This register is accessed only upon hard reset,
     // I don't know what's it supposed to be.
     return;
   }
  
   assert( reg == ATA_DATA );
   assert( direction == currentDirection );
   assert( numBytes == 0x200 );
   assert( pendingDataCount == 0x100 );
   pendingDataCount = 0; // take care of the whole sector at once;
   if (direction == 1) {
     memcpy( ideBuffer, memBuff, 0x200 );
     for( int i = 0; i < 0x100; i++ )
       ideBuffer[i] = swapBytes( ideBuffer[i] );
     flushSector();
   } else {
     for( int i = 0; i < 0x100; i++ )
       ideBuffer[i] = swapBytes( ideBuffer[i] );
     memcpy( memBuff, ideBuffer, 0x200 );
     bufferNextSector();
   }
}

u16 IDEController::readReg( u32 reg ) {
   // a send form IDE's point of view
   u16 value = registers[reg];
   switch ( reg ) {
   case ATA_ALT_STATUS:
     if( pendingDataCount > 0 || pendingSectors > 0 ) {
       value = STATUS_DRQ;
     }
     else {
       value = STATUS_RDY;
     }
     break;
      case ATA_DATA:
         assert( pendingDataCount > 0 );
         assert( pendingDataCount <= 0x100 );
         value = ideBuffer[0x100 - pendingDataCount];
         pendingDataCount--;
         if( pendingDataCount == 0 )
            bufferNextSector(  );
         break;
      case ATA_STATUS:
         if( pendingDataCount > 0 || pendingSectors > 0 ) {
            value = STATUS_DRQ;
         }
         else {
            value = STATUS_RDY;
         }
         break;
   case ATA_ERROR:
   case ATA_NSECTOR:
   case ATA_SECTOR:
   case ATA_LCYL:
   case ATA_HCYL:
   case ATA_SELECT:
   case ATA_UNKNOWN:
     break;
   default:
     console.output("Attempt to read IDE register: %x", reg);
     assert(false);
      }
   if( TRACE_IDE )
      gmini->tracer.log( "IDE REG read:  %2x -> %4x", reg, value );
   return value;
}

void IDEController::bufferNextSector(  ) {
   if( pendingSectors > 0 ) {
      readSector( nextSector, ( u8 * ) ideBuffer );
      pendingDataCount = 0x100;
      pendingSectors--;
      nextSector++;
   }
}

void IDEController::flushSector(  ) {
    writeSector( nextSector, ( u8 * ) ideBuffer );
    if (pendingSectors > 0) {
        pendingDataCount = 0x100;
        pendingSectors--;
        nextSector++;
    } else
        if( TRACE_SECT )
            gmini->tracer.log( "IDE SECT write over");
}


void IDEController::writeReg( u32 reg, u16 value ) {
   // a receive from IDE's point of view

   if( TRACE_IDE )
       gmini->tracer.log( "IDE REG write: %2x <- %4x", reg, value );
   switch ( reg ) {
   case ATA_ERROR:
   case ATA_NSECTOR:
   case ATA_SECTOR:
   case ATA_LCYL:
   case ATA_HCYL:
   case ATA_SELECT:
           break;
       case ATA_DATA:
           assert( pendingDataCount > 0 );
           assert( pendingDataCount <= 0x100 );
           
           ideBuffer[0x100 - pendingDataCount] = value;
           pendingDataCount--;
           if( pendingDataCount == 0 )
               flushSector(  );           
           break;
       case ATA_COMMAND:
         if( value == CMD_IDENTIFY ) {
            // start read config data
            pendingDataCount = 0x100;
            memcpy( ideBuffer, configData, 0x200 );
         }
         else if( value == CMD_READ_SECTORS || value == CMD_WRITE_SECTORS ||
                  value == CMD_READ_MULTIPLE || value == CMD_WRITE_MULTIPLE) {
             if (value == CMD_READ_SECTORS || value == CMD_WRITE_SECTORS)
                 currentDirection = ( value & 0x10 ) >> 4;
             else
                 currentDirection = value & 1;
            nextSector =
              ( ( registers[ATA_SELECT] & 0xF ) << 24 ) + ( registers[ATA_HCYL] << 16 ) +
              ( registers[ATA_LCYL] << 8 ) + ( registers[ATA_SECTOR] );
            pendingSectors = registers[ATA_NSECTOR];
            if( TRACE_SECT )
               gmini->tracer.log( "IDE SECT op: lba=%x n=%x dir=%d",
                                  nextSector, pendingSectors, currentDirection );

            if( currentDirection == 0 )
               bufferNextSector(  );
            else {
                pendingDataCount = 0x100;
                pendingSectors--;
            }
         }

         break;
      case ATA_CONTROL:
         if( value == 0x0E ) {
            // probably reset 

            // soft reset
            // Sp 8 6 <- D4
            // Sp A C <- D6
            // 13 12 <- D8
            // 5 4 <- da
            // sp 10 e <- de
            registers[ATA_NSECTOR] = 1;
            registers[ATA_SECTOR] = 1;
            registers[0xDE] = 0; // bit 0 must be zero... But the rest ???

            // either: (for CDM)
            // registers[ATA_LCYL] = 0x14; 
            // registers[ATA_HCYL] = 0xeb;

            // or: (for HDM) (causes the firmware to do more initialisations )
            registers[ATA_LCYL] = 0;
            registers[ATA_HCYL] = 0;
         }
         break;
   default:
     console.output("Attempt to write IDE register: %x", reg);
     assert(false);
   }
   registers[reg] = value;

}
