/*  Project: Open Source Gmini Tools
    file: SymMap.cpp, mantains a dynamic symbol mapping

    Copyright (C) 2004  The gmEmu Project <gmemu@donat.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <fstream>
#include <stdio.h>
#include "SymMap.h"
#include "Console.h"

SymMap::SymMap(  ) {
   load( "std/symbols.std" );
}

SymMap::~SymMap(  ) {
   map < u32, Reference * >::const_iterator itr = symmap.begin(  );
   while( itr != symmap.end(  ) ) {
      delete itr->second;
      ++itr;
   }
}

int SymMap::load( const char *mapfile ) {
   char buffer[BUFLEN];
   if( !mapfile )
      return 0;
   ifstream data( mapfile, ios::in );
   if( !data ) {
      console.output( "Could not open map file: %s\n", mapfile );
      return -1;
   }
   console.output( "Loading symbols from %s: ", mapfile );
   char symname[BUFLEN];
   char refsym[BUFLEN];
   char rtypename[BUFLEN];
   u32 addr = 0;
   data.getline( buffer, BUFLEN );
   int counter = 0;
   while( !data.eof(  ) ) {
      if( strncmp( "0x", buffer, 2 ) == 0 ) {
         sscanf( buffer, "0x%X %s", &addr, symname );
         if( symname[0] != '|' ) {
            Reference *ref = new Reference( symname );
            symmap[addr] = ref;
         }
         else {                 // "|module$$segment$$type|"
            char *str2 = strchr( symname, '$' ) + 2;
            char *str3 = strchr( str2, '$' ) + 2;
            int length1 = str2 - symname - 3;
            int length2 = str3 - str2 - 2;
            int length3 = strlen( str3 ) - 1;
            char *module = new char[length1];
            char *segment = new char[length2];
            char *datatype = new char[length3];
            strncpy( module, symname + 1, length1 );
            module[length1] = '\0';
            strncpy( segment, str2, length2 );
            segment[length2] = '\0';
            strncpy( datatype, str3, length3 );
            datatype[length3] = '\0';
            console.output( "|%s$$%s$$%s| = 0x%x\n", module, segment, datatype, addr );
         }
         ++counter;
      }
      else if( buffer[0] == ' ' ) { // References (line begins with single
         // space) " blah 0xXXXXXX\n"
         u32 from;
         sscanf( buffer, "%s 0x%X", rtypename, &from );
         int index = 0;
         while( index < 4 && strcmp( Reference::typenames[index], rtypename ) != 0 )
            ++index;
         if( index < 4 )
            addref( from, addr, ( Reference::access ) index );
      }
      else if( strchr( buffer, '+' ) != NULL ) {
         // Symbol defined by offset of another
         // referenced symbol must be defined earlier in the file
         // only positive offsets allowed
         // "symbol+0xXXXX Symbol"
         char *plus = strchr( buffer, '+' );
         strncpy( refsym, buffer, plus - buffer );
         refsym[plus - buffer] = '\0';
         sscanf( plus + 1, "0x%x", &addr );
         strcpy( symname, strchr( buffer, ' ' ) + 1 );
         addr += addrof( refsym );
         if( addrof( refsym ) != 0xFFFFFFF0 ) {
            symmap[addr] = new Reference( symname );
            ++counter;
         }
      }
      else if( strncmp( "DATASEGMENT", buffer, 11 ) == 0 ) {
         // data segment description "DATASEGMENT start length"
         // if a range falls outside a segment description it is assumed to be code.
         u32 start;
         u32 length;
         sscanf( buffer, "DATASEGMENT 0x%X 0x%X", &start, &length );
         setData( start, length );
      }

      data.getline( buffer, BUFLEN );
   }
   console.output( "%d symbols loaded.\n", counter );
   data.close(  );
   return 0;
}

const char *SymMap::get( u32 addr ) {
   if( symmap.find( addr ) != symmap.end(  ) )
      return symmap[addr]->getname(  );
   return NULL;
}

const Reference *SymMap::getReference( u32 addr ) {
   if( symmap.find( addr ) != symmap.end(  ) )
      return symmap[addr];
   return NULL;

}

void SymMap::addref( u32 from, u32 to, Reference::access type ) {

   if( symmap.find( to ) == symmap.end(  ) ) {
      char name[24];
      switch ( type ) {
         case Reference::DATA:
            sprintf( name, "data_%06X", to );
            break;
         case Reference::RJUMP:
            sprintf( name, "rel_%06X", to );
            break;
         case Reference::AJUMP:
            sprintf( name, "abs_%06X", to );
            break;
         case Reference::CALL:
            sprintf( name, "sub_%06X", to );
            break;
      }
      symmap[to] = new Reference( name );
   }
   if( symmap.find( to ) != symmap.end(  ) )
      symmap[to]->add( from, type );
}

const char *SymMap::getref( u32 addr, char *buffer, bool html ) {
   if( symmap.find( addr ) != symmap.end(  ) ) {
      if( symmap[addr]->getrefs( buffer, html ) == 0 )
         return NULL;
      return buffer;
   }
   return NULL;
}

const u32 SymMap::addrof( char *symbol ) {
   map < u32, Reference * >::const_iterator itr = symmap.begin(  );
   while( itr != symmap.end(  ) && ( strcmp( symbol, itr->second->getname(  ) ) != 0 ) ) {
      ++itr;
   }
   if( itr != symmap.end(  ) )
      return itr->first;
   return 0xFFFFFFF0;
}

static int commonPrefix(const char* s1, const char* s2) {
  int i = 0;
  while (s1[i] && s2[i] && s1[i] == s2[i]) {
    i++;
  }
  return i;
}

const u22 SymMap::resolveInteractive( char *symbol ) {
  u22 res = 0xFFFFFFF0;
  int best = 3;
  int len = strlen(symbol);
  for (map < u32, Reference * >::const_iterator itr = symmap.begin(  );
       itr != symmap.end(  ); ++itr) {
    const char* sname = itr->second->getname(  );
    int com = commonPrefix(symbol, sname);
    //printf("%s -> %d\n", itr->second->getname(  ), len);
    if (com == strlen(sname) && com == len)
      return itr->first;
    if (com > best) {
      best = com;
      res = itr->first;
    }
  }
  return res;
}


const bool SymMap::isData( u32 addr ) {
   map<u32,u32>::iterator itr = datamap.begin();
   while( itr != datamap.end() ) {
      if( addr >= itr->first && addr < itr->second )
         return true;
      ++itr;
   }
   return false;
}

void SymMap::setData( u32 start, u32 length ) {
   u32 end = start + length;
   bool done;
   do {
      done = true;
      map<u32,u32>::iterator itr = datamap.begin();
      while( itr != datamap.end() ) {
         bool lap1 = false;
         bool lap2 = false;
         if( start < itr->first && end > itr->first )
            lap1 = true;
         if( start < itr->second && end > itr->second )
            lap2 = true;
         if( lap1 || lap2 ) {
            if( lap1 && !lap2 )
               end = itr->second;
            if( !lap1 && lap2 )
               start = itr->first;
            datamap.erase( itr->first );
            done = false;
            break;
         }
         else if( start > itr->first && end < itr->second ) {
            length = 0;
            break;
         }
         ++itr;
      }
   } while( !done );
   if( length != 0 )
      datamap[ start ] = end;
}

const int SymMap::dump( const char *file ) {
   ofstream data( file, ios::out );
   if( !data ) {
      console.output( "Could not open map file: %s\n", file );
      return -1;
   }
   char *buffer = new char[128 * 1024];

   map<u32,u32>::iterator ditr = datamap.begin();
   while( ditr != datamap.end() ) {
      int length = sprintf( buffer, "DATASEGMENT 0x%06X 0x%06X\n", ditr->first, ditr->second - ditr->first );
      data.write( buffer, length );
      ++ditr;
   }
   map < u32, Reference * >::const_iterator itr = symmap.begin(  );
   while( itr != symmap.end(  ) ) {
      int length = sprintf( buffer, "0x%06X %s\n", itr->first, itr->second->getname(  ) );
      length += itr->second->getrefs( buffer + length );
      data.write( buffer, length );
      ++itr;
   }
   delete[] buffer;
   data.close(  );
   return 0;
}
